#ifndef KR2_KORD_STATS_H
#define KR2_KORD_STATS_H

#include <stdio.h>
#include <cstddef>
#include <vector>
#include <algorithm>
#include <iterator>
#include <numeric>
#include <cmath>
#include <mutex>
#include <boost/interprocess/sync/scoped_lock.hpp>

#include <iostream>

struct Ana {
    double min = -1.0;
    double max = -1.0;
    double avg = -1.0;
    double std = -1.0;

    void set(const Ana &a_a) {
        this->min = a_a.min;
        this->max = a_a.max;
        this->avg = a_a.avg;
        this->std = a_a.std;
    }

    void reset() {
        this->min = 0.0;
        this->max = 0.0;
        this->avg = 0.0;
        this->std = 0.0;
    }
};


template<class T> class Accum {

    public:
        /// Allocate/Deallocate the vector automatically
        Accum(std::size_t N) : N(N), _data(N,T{}) {};

    public:

        void measurement(T x) {
            if (_idx >= N) {_idx = 0; _fulled = true; }
            _data[_idx] = x;
            _pidx = _idx;
            _idx++;
        }
        // get Vector
        std::vector<T> toVector(){
            if (!this->_fulled) { return this->slicing(this->_data, 0, this->_idx); } else { return this->_data; };
        }
        
        void compute() {
            
            std::vector<T> _v;
            
            if (!this->_fulled) { _v = this->slicing(this->_data, 0, this->_idx); }
            else                { _v = this->_data; }

            // printResult(_v);
            
            auto _bidx = _v.begin();
            auto _eidx = _v.end();
            auto _size = _v.size();

            this->_min = *min_element(_bidx, _eidx);
            this->_max = *max_element(_bidx, _eidx);
            
            if(_data.empty()) { this->_avg = 0.0; this->_std = 0.0; }
            else
            {
                auto const count = static_cast<double>(_size);
                this->_avg = std::accumulate(_bidx, _eidx, 0.0) / count;

                double accum = 0.0;
                std::for_each (_bidx, _eidx, [&](const double x) {
                    accum += (x - this->_avg) * (x - this->_avg);
                });
                this->_std = std::sqrt(accum / (count-1));
            }

            // fill the Ana struct
            this -> _a.min = this->_min;
            this -> _a.max = this->_max;
            this -> _a.avg = this->_avg;
            this -> _a.std = this->_std;
        }

        T get_min()      { return this->_min; }
        T get_max()      { return this->_max; }
        double get_avg() { if (this->_fulled) return this->_avg; else return 0.0;}
        double get_std() { return this->_std; }

        Ana get_ana()    { return this->_a;   }

        T get_last_measurement() { return _data[_pidx]; }

        void reset() {
            _data.assign(N, 0);
            _idx = 0;
            _pidx = 0;
            _fulled = false;
            _min = 0;
            _max = 0;
            _avg = 0;
            _a = Ana();
            // if (!this->_data.empty()) 
            //     this->_data.clear(); 
        }

    private:

        std::size_t N = 1000;
        std::vector<T> _data;

        std::vector<int>::size_type _idx  = 0;
        std::vector<int>::size_type _pidx = 0;
        T _min           = _data[0];
        T _max           = _data[0];
        double _avg      = 0.0;
        double _std      = 0.0;
        bool _fulled     = false;
        Ana _a           = Ana();
        
        std::vector<T> slicing(std::vector<T>& arr, int S, int E)
        {
            // Starting and Ending iterators
            auto start = arr.begin() + S;
            auto end   = arr.begin() + E;
        
            // To store the sliced vector
            std::vector<T> result(E - S);
        
            // Copy vector using copy function()
            copy(start, end, result.begin());
        
            // Return the final sliced vector
            return result;
        }

    // Function to print the vector ans
    void printResult(const std::vector<T>& ans)
    {
    
        // Traverse the vector ans
        for (auto& it : ans) {
    
            // Print elements
            std::cout << it << ' ';
        }
        std::cout << std::endl;
    }
};

#endif